# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: gravity/v1/gravity.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'gravity/v1/gravity.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18gravity/v1/gravity.proto\x12\ngravity.v1\x1a\x1fgoogle/protobuf/timestamp.proto\"\xab\x02\n\x07\x43rawler\x12\x12\n\ncrawler_id\x18\x01 \x01(\t\x12-\n\x08\x63riteria\x18\x02 \x01(\x0b\x32\x1b.gravity.v1.CrawlerCriteria\x12.\n\nstart_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x37\n\x13\x64\x65registration_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x30\n\x0c\x61rchive_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\'\n\x05state\x18\x06 \x01(\x0b\x32\x18.gravity.v1.CrawlerState\x12\x19\n\x11\x64\x61taset_workflows\x18\x07 \x03(\t\"w\n\x0f\x43rawlerCriteria\x12\x10\n\x08platform\x18\x01 \x01(\t\x12\r\n\x05topic\x18\x02 \x01(\t\x12\x35\n\x0cnotification\x18\x03 \x01(\x0b\x32\x1f.gravity.v1.CrawlerNotification\x12\x0c\n\x04mock\x18\x04 \x01(\x08\"/\n\x13\x43rawlerNotification\x12\n\n\x02to\x18\x01 \x01(\t\x12\x0c\n\x04link\x18\x02 \x01(\t\"C\n\x06HfRepo\x12\x11\n\trepo_name\x18\x01 \x01(\t\x12\x11\n\trow_count\x18\x02 \x01(\x04\x12\x13\n\x0blast_update\x18\x03 \x01(\t\"u\n\x0c\x43rawlerState\x12\x0e\n\x06status\x18\x01 \x01(\t\x12\x17\n\x0f\x62ytes_collected\x18\x02 \x01(\x04\x12\x19\n\x11records_collected\x18\x03 \x01(\x04\x12!\n\x05repos\x18\x04 \x03(\x0b\x32\x12.gravity.v1.HfRepo\"\xbe\x01\n\x10GravityTaskState\x12\x17\n\x0fgravity_task_id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0e\n\x06status\x18\x03 \x01(\t\x12.\n\nstart_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x13\n\x0b\x63rawler_ids\x18\x05 \x03(\t\x12.\n\x11\x63rawler_workflows\x18\x06 \x03(\x0b\x32\x13.gravity.v1.Crawler\"~\n\x16GetGravityTasksRequest\x12\x1c\n\x0fgravity_task_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x1d\n\x10include_crawlers\x18\x02 \x01(\x08H\x01\x88\x01\x01\x42\x12\n\x10_gravity_task_idB\x13\n\x11_include_crawlers\"T\n\x17GetGravityTasksResponse\x12\x39\n\x13gravity_task_states\x18\x01 \x03(\x0b\x32\x1c.gravity.v1.GravityTaskState\".\n\x0bGravityTask\x12\r\n\x05topic\x18\x01 \x01(\t\x12\x10\n\x08platform\x18\x02 \x01(\t\"`\n\x13NotificationRequest\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x12\x19\n\x0credirect_url\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\x0f\n\r_redirect_url\"\'\n\x11GetCrawlerRequest\x12\x12\n\ncrawler_id\x18\x01 \x01(\t\":\n\x12GetCrawlerResponse\x12$\n\x07\x63rawler\x18\x01 \x01(\x0b\x32\x13.gravity.v1.Crawler\"\xca\x01\n\x18\x43reateGravityTaskRequest\x12.\n\rgravity_tasks\x18\x01 \x03(\x0b\x32\x17.gravity.v1.GravityTask\x12\x0c\n\x04name\x18\x02 \x01(\t\x12>\n\x15notification_requests\x18\x03 \x03(\x0b\x32\x1f.gravity.v1.NotificationRequest\x12\x1c\n\x0fgravity_task_id\x18\x04 \x01(\tH\x00\x88\x01\x01\x42\x12\n\x10_gravity_task_id\"4\n\x19\x43reateGravityTaskResponse\x12\x17\n\x0fgravity_task_id\x18\x01 \x01(\t\"{\n\x13\x42uildDatasetRequest\x12\x12\n\ncrawler_id\x18\x01 \x01(\t\x12>\n\x15notification_requests\x18\x02 \x03(\x0b\x32\x1f.gravity.v1.NotificationRequest\x12\x10\n\x08max_rows\x18\x03 \x01(\x03\"P\n\x14\x42uildDatasetResponse\x12\x12\n\ndataset_id\x18\x01 \x01(\t\x12$\n\x07\x64\x61taset\x18\x02 \x01(\x0b\x32\x13.gravity.v1.Dataset\"=\n\x06Nebula\x12\r\n\x05\x65rror\x18\x01 \x01(\t\x12\x17\n\x0f\x66ile_size_bytes\x18\x02 \x01(\x03\x12\x0b\n\x03url\x18\x03 \x01(\t\"\xb9\x02\n\x07\x44\x61taset\x12\x1b\n\x13\x63rawler_workflow_id\x18\x01 \x01(\t\x12/\n\x0b\x63reate_date\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0b\x65xpire_date\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12&\n\x05\x66iles\x18\x04 \x03(\x0b\x32\x17.gravity.v1.DatasetFile\x12\x0e\n\x06status\x18\x05 \x01(\t\x12\x16\n\x0estatus_message\x18\x06 \x01(\t\x12&\n\x05steps\x18\x07 \x03(\x0b\x32\x17.gravity.v1.DatasetStep\x12\x13\n\x0btotal_steps\x18\x08 \x01(\x03\x12\"\n\x06nebula\x18\t \x01(\x0b\x32\x12.gravity.v1.Nebula\"\x9b\x01\n\x0b\x44\x61tasetFile\x12\x11\n\tfile_name\x18\x01 \x01(\t\x12\x17\n\x0f\x66ile_size_bytes\x18\x02 \x01(\x04\x12\x31\n\rlast_modified\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08num_rows\x18\x04 \x01(\x04\x12\x0e\n\x06s3_key\x18\x05 \x01(\t\x12\x0b\n\x03url\x18\x06 \x01(\t\"@\n\x0b\x44\x61tasetStep\x12\x10\n\x08progress\x18\x01 \x01(\x01\x12\x0c\n\x04step\x18\x02 \x01(\x03\x12\x11\n\tstep_name\x18\x03 \x01(\t\"\'\n\x11GetDatasetRequest\x12\x12\n\ndataset_id\x18\x01 \x01(\t\":\n\x12GetDatasetResponse\x12$\n\x07\x64\x61taset\x18\x01 \x01(\x0b\x32\x13.gravity.v1.Dataset\"3\n\x18\x43\x61ncelGravityTaskRequest\x12\x17\n\x0fgravity_task_id\x18\x01 \x01(\t\",\n\x19\x43\x61ncelGravityTaskResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"*\n\x14\x43\x61ncelDatasetRequest\x12\x12\n\ndataset_id\x18\x01 \x01(\t\"(\n\x15\x43\x61ncelDatasetResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"X\n\x1f\x44\x61tasetBillingCorrectionRequest\x12\x1b\n\x13requested_row_count\x18\x01 \x01(\x03\x12\x18\n\x10\x61\x63tual_row_count\x18\x02 \x01(\x03\"9\n DatasetBillingCorrectionResponse\x12\x15\n\rrefund_amount\x18\x01 \x01(\x01\x32\xea\x05\n\x0eGravityService\x12Z\n\x0fGetGravityTasks\x12\".gravity.v1.GetGravityTasksRequest\x1a#.gravity.v1.GetGravityTasksResponse\x12K\n\nGetCrawler\x12\x1d.gravity.v1.GetCrawlerRequest\x1a\x1e.gravity.v1.GetCrawlerResponse\x12`\n\x11\x43reateGravityTask\x12$.gravity.v1.CreateGravityTaskRequest\x1a%.gravity.v1.CreateGravityTaskResponse\x12Q\n\x0c\x42uildDataset\x12\x1f.gravity.v1.BuildDatasetRequest\x1a .gravity.v1.BuildDatasetResponse\x12K\n\nGetDataset\x12\x1d.gravity.v1.GetDatasetRequest\x1a\x1e.gravity.v1.GetDatasetResponse\x12`\n\x11\x43\x61ncelGravityTask\x12$.gravity.v1.CancelGravityTaskRequest\x1a%.gravity.v1.CancelGravityTaskResponse\x12T\n\rCancelDataset\x12 .gravity.v1.CancelDatasetRequest\x1a!.gravity.v1.CancelDatasetResponse\x12u\n\x18\x44\x61tasetBillingCorrection\x12+.gravity.v1.DatasetBillingCorrectionRequest\x1a,.gravity.v1.DatasetBillingCorrectionResponseB4Z2macrocosm-os/rift/constellation_api/gen/gravity/v1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'gravity.v1.gravity_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z2macrocosm-os/rift/constellation_api/gen/gravity/v1'
  _globals['_CRAWLER']._serialized_start=74
  _globals['_CRAWLER']._serialized_end=373
  _globals['_CRAWLERCRITERIA']._serialized_start=375
  _globals['_CRAWLERCRITERIA']._serialized_end=494
  _globals['_CRAWLERNOTIFICATION']._serialized_start=496
  _globals['_CRAWLERNOTIFICATION']._serialized_end=543
  _globals['_HFREPO']._serialized_start=545
  _globals['_HFREPO']._serialized_end=612
  _globals['_CRAWLERSTATE']._serialized_start=614
  _globals['_CRAWLERSTATE']._serialized_end=731
  _globals['_GRAVITYTASKSTATE']._serialized_start=734
  _globals['_GRAVITYTASKSTATE']._serialized_end=924
  _globals['_GETGRAVITYTASKSREQUEST']._serialized_start=926
  _globals['_GETGRAVITYTASKSREQUEST']._serialized_end=1052
  _globals['_GETGRAVITYTASKSRESPONSE']._serialized_start=1054
  _globals['_GETGRAVITYTASKSRESPONSE']._serialized_end=1138
  _globals['_GRAVITYTASK']._serialized_start=1140
  _globals['_GRAVITYTASK']._serialized_end=1186
  _globals['_NOTIFICATIONREQUEST']._serialized_start=1188
  _globals['_NOTIFICATIONREQUEST']._serialized_end=1284
  _globals['_GETCRAWLERREQUEST']._serialized_start=1286
  _globals['_GETCRAWLERREQUEST']._serialized_end=1325
  _globals['_GETCRAWLERRESPONSE']._serialized_start=1327
  _globals['_GETCRAWLERRESPONSE']._serialized_end=1385
  _globals['_CREATEGRAVITYTASKREQUEST']._serialized_start=1388
  _globals['_CREATEGRAVITYTASKREQUEST']._serialized_end=1590
  _globals['_CREATEGRAVITYTASKRESPONSE']._serialized_start=1592
  _globals['_CREATEGRAVITYTASKRESPONSE']._serialized_end=1644
  _globals['_BUILDDATASETREQUEST']._serialized_start=1646
  _globals['_BUILDDATASETREQUEST']._serialized_end=1769
  _globals['_BUILDDATASETRESPONSE']._serialized_start=1771
  _globals['_BUILDDATASETRESPONSE']._serialized_end=1851
  _globals['_NEBULA']._serialized_start=1853
  _globals['_NEBULA']._serialized_end=1914
  _globals['_DATASET']._serialized_start=1917
  _globals['_DATASET']._serialized_end=2230
  _globals['_DATASETFILE']._serialized_start=2233
  _globals['_DATASETFILE']._serialized_end=2388
  _globals['_DATASETSTEP']._serialized_start=2390
  _globals['_DATASETSTEP']._serialized_end=2454
  _globals['_GETDATASETREQUEST']._serialized_start=2456
  _globals['_GETDATASETREQUEST']._serialized_end=2495
  _globals['_GETDATASETRESPONSE']._serialized_start=2497
  _globals['_GETDATASETRESPONSE']._serialized_end=2555
  _globals['_CANCELGRAVITYTASKREQUEST']._serialized_start=2557
  _globals['_CANCELGRAVITYTASKREQUEST']._serialized_end=2608
  _globals['_CANCELGRAVITYTASKRESPONSE']._serialized_start=2610
  _globals['_CANCELGRAVITYTASKRESPONSE']._serialized_end=2654
  _globals['_CANCELDATASETREQUEST']._serialized_start=2656
  _globals['_CANCELDATASETREQUEST']._serialized_end=2698
  _globals['_CANCELDATASETRESPONSE']._serialized_start=2700
  _globals['_CANCELDATASETRESPONSE']._serialized_end=2740
  _globals['_DATASETBILLINGCORRECTIONREQUEST']._serialized_start=2742
  _globals['_DATASETBILLINGCORRECTIONREQUEST']._serialized_end=2830
  _globals['_DATASETBILLINGCORRECTIONRESPONSE']._serialized_start=2832
  _globals['_DATASETBILLINGCORRECTIONRESPONSE']._serialized_end=2889
  _globals['_GRAVITYSERVICE']._serialized_start=2892
  _globals['_GRAVITYSERVICE']._serialized_end=3638
# @@protoc_insertion_point(module_scope)



from typing import AsyncIterator, Generic, TypeVar

from grpc.aio import Channel

T = TypeVar("T")


class StreamGenerator(Generic[T], AsyncIterator[T]):
    """Generator for async iterator to close the channel when the iterator is closed."""

    def __init__(self, stream_response: AsyncIterator[T], channel: Channel):
        self._generator = self.stream_generator(stream_response)
        self._channel = channel

    async def stream_generator(self, stream_response: AsyncIterator[T]):
        async for chunk in stream_response:
            yield chunk

    def __aiter__(self):
        return self

    async def __anext__(self) -> T:
        try:
            return await self._generator.__anext__()
        except StopAsyncIteration:
            await self._channel.close()
            raise

    async def aclose(self):
        await self._channel.close()



import asyncio
from typing import List, Optional
import grpc
from google.protobuf.json_format import MessageToDict
from macrocosmos import __package_name__, __version__
from macrocosmos.generated.sn13.v1 import sn13_validator_pb2, sn13_validator_pb2_grpc
from macrocosmos.resources._client import BaseClient
from macrocosmos.types import MacrocosmosError


class AsyncSn13:
    """Asynchronous SN13 resource for the Data Universe (subnet 13) API on Bittensor."""

    def __init__(self, client: BaseClient):
        """
        Initialize the asynchronous SN13 resource.

        Args:
            client: The client to use for the resource.
        """
        self._client = client

    async def OnDemandData(
        self,
        source: str,
        usernames: List[str],
        keywords: List[str],
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        limit: int = 5,
    ) -> sn13_validator_pb2.OnDemandDataResponse:
        """
        Retrieves on-demand data from the SN13 API service asynchronously, based on the provided parameters.

        Args:
            source (str): The data source (X or Reddit)
            usernames (List[str]): List of usernames to fetch data from
            keywords (List[str]): List of keywords to search for
            start_date (str): Date from which we want to start fetching data. ISO 8601 formatted date string (e.g. "2024-01-01T00:00:00Z")
            end_date (str): Date up to which we want to fetch data. ISO 8601 formatted date string (e.g. "2024-01-01T00:00:00Z")
            limit (int): Maximum number of results to return

        Returns:
            dict:
                - status (str): The request status
                - data (List[dict]): The data object returned by the miners
                - meta (dict): Additional metadata about the request
        """
        request = sn13_validator_pb2.OnDemandDataRequest(
            source=source,
            usernames=usernames,
            keywords=keywords,
            start_date=start_date,
            end_date=end_date,
            limit=limit,
        )

        return await self._make_request("OnDemandData", request)

    async def _make_request(self, method_name, request):
        """
        Make a request to the SN13 service.

        Args:
            method_name: The name of the method to call.
            request: The request message.

        Returns:
            The response from the service.
        """
        metadata = [
            ("x-source", self._client.app_name),
            ("x-client-id", __package_name__),
            ("x-client-version", __version__),
            ("authorization", f"Bearer {self._client.api_key}"),
            (
                "remote-user-email",
                "example@example.com",
            ),  # TODO: Remove this once authentication is fully implemented
        ]

        compression = grpc.Compression.Gzip if self._client.compress else None

        retries = 0
        last_error = None
        while retries <= self._client.max_retries:
            try:
                channel = self._client.get_channel()
                stub = sn13_validator_pb2_grpc.Sn13ServiceStub(channel)
                method = getattr(stub, method_name)
                response = await method(
                    request,
                    metadata=metadata,
                    timeout=self._client.timeout,
                    compression=compression,
                )
                await channel.close()
                # MessageToDict removes verbosity due to data and meta fields being google.protobuf.Struct types
                return MessageToDict(
                    response, preserving_proto_field_name=True
                )  # preserving_proto_field_name=True removes lowerCamelCase formatting
            except grpc.RpcError as e:
                last_error = MacrocosmosError(f"RPC error: {e.code()}: {e.details()}")
                retries += 1
                await channel.close()
            except Exception as e:
                await channel.close()
                raise MacrocosmosError(f"Error calling {method_name}: {e}")

        raise last_error


class SyncSn13:
    """Synchronous SN13 resource for the Data Universe (subnet 13) API on Bittensor."""

    def __init__(self, client: BaseClient):
        """
        Initialize the synchronous SN13 resource.

        Args:
            client: The client to use for the resource.
        """
        self._client = client
        self._async_sn13 = AsyncSn13(client)

    def OnDemandData(
        self,
        source: str,
        usernames: List[str],
        keywords: List[str],
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        limit: int = 5,
    ) -> sn13_validator_pb2.OnDemandDataResponse:
        """
        Retrieves on-demand data from the SN13 API service synchronously, based on the provided parameters.

        Args:
            source (str): The data source (X or Reddit)
            usernames (List[str]): List of usernames to fetch data from
            keywords (List[str]): List of keywords to search for
            start_date (str): Date from which we want to start fetching data. ISO 8601 formatted date string (e.g. "2024-01-01T00:00:00Z")
            end_date (str): Date up to which we want to fetch data. ISO 8601 formatted date string (e.g. "2024-01-01T00:00:00Z")
            limit (int): Maximum number of results to return

        Returns:
            dict:
                - status (str): The request status
                - data (List[dict]): The data object returned by the miners
                - meta (dict): Additional metadata about the request
        """
        return asyncio.run(
            self._async_sn13.OnDemandData(
                source=source,
                usernames=usernames,
                keywords=keywords,
                start_date=start_date,
                end_date=end_date,
                limit=limit,
            )
        )



from abc import ABC
import grpc
from typing import Optional
import os
from macrocosmos.types import MacrocosmosError

DEFAULT_BASE_URL = "constellation.api.cloud.macrocosmos.ai"
DEFAULT_USE_HTTPS = True


class BaseClient(ABC):
    """
    Abstract base class for client.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        timeout: Optional[int] = None,
        max_retries: int = 0,
        secure: Optional[bool] = None,
        compress: bool = True,
        app_name: Optional[str] = None,
    ):
        """
        Initialize the abstract base class for the client.

        Args:
            api_key: The API key.
            base_url: The base URL for the API.
            timeout: Time to wait for a response in seconds. (default: None)
            max_retries: The maximum number of retries. (default: 0)
            secure: Whether to use HTTPS. Set this if you're using a custom base URL.
            compress: Whether to compress the request using gzip (default: True).
            app_name: The name of the application using the client.
        """
        if not api_key:
            api_key = os.environ.get("MACROCOSMOS_API_KEY")
            api_key = api_key.strip() if api_key else ""
            if not api_key:
                raise MacrocosmosError(
                    "The api_key client option must be set either by passing api_key to the client or by setting the MACROCOSMOS_API_KEY environment variable"
                )

        if not base_url:
            base_url = os.environ.get("MACROCOSMOS_BASE_URL")
            base_url = base_url.strip() if base_url else ""
            if not base_url:
                base_url = DEFAULT_BASE_URL

        if not isinstance(secure, bool):
            secure_str = os.environ.get("MACROCOSMOS_USE_HTTPS", "")
            secure_str = secure_str.strip().lower() if secure_str else ""
            if secure_str == "":
                secure = DEFAULT_USE_HTTPS
            else:
                secure = secure_str in ["true", "1", "yes", "y", "on"]

        self.api_key = api_key
        self.base_url = base_url.rstrip("/")
        self.timeout = timeout
        self.max_retries = max_retries
        self.secure = secure
        self.compress = compress
        self.app_name = app_name

    def get_channel(self) -> grpc.aio.Channel:
        """
        Get a channel for the given client.
        """
        if self.secure:
            return grpc.aio.secure_channel(
                self.base_url, grpc.ssl_channel_credentials()
            )
        return grpc.aio.insecure_channel(self.base_url)



"""
NOTE: This demo is showing how the raw gRPC endpoint works using the protobufs.

This is not recommended for most use cases, as the macrocosmos client handles
compression, retries, and other details for you.  You should use the client
interface instead.  See the other examples for how to use the client.
"""

import os

import grpc

from macrocosmos import __package_name__, __version__
from macrocosmos.resources._client import DEFAULT_BASE_URL
from macrocosmos.generated.apex.v1 import apex_pb2, apex_pb2_grpc


def demo_chat_completion():
    """Demo processing a single chat completion using the raw gRPC API."""

    channel = grpc.secure_channel(DEFAULT_BASE_URL, grpc.ssl_channel_credentials())
    stub = apex_pb2_grpc.ApexServiceStub(channel)

    messages = [apex_pb2.ChatMessage(role="user", content="Hello, how are you today?")]

    sampling_params = apex_pb2.SamplingParameters(
        temperature=0.7,
        top_p=0.9,
        max_new_tokens=256,
        do_sample=True,
    )

    api_key = os.environ.get("APEX_API_KEY", os.environ.get("MACROCOSMOS_API_KEY"))
    metadata = [
        ("x-source", "examples/apex_chat_grpc"),
        ("x-client-id", __package_name__),
        ("x-client-version", __version__),
        ("authorization", f"Bearer {api_key}"),
    ]

    request = apex_pb2.ChatCompletionRequest(
        messages=messages,
        sampling_parameters=sampling_params,
        stream=False,
    )

    try:
        response = stub.ChatCompletion(
            request,
            metadata=metadata,
            timeout=30.0,
            compression=grpc.Compression.Gzip,
        )

        print("Response received:")
        print(f"ID: {response.id}")

        for i, choice in enumerate(response.choices):
            print(f"\nChoice {i}:")
            print(f"  Finish reason: {choice.finish_reason}")
            print(f"  Message content: {choice.message.content}")
            if choice.message.refusal:
                print(f"  Refusal: {choice.message.refusal}")

    except grpc.RpcError as e:
        print(f"RPC error: {e.code()}: {e.details()}")


def demo_chat_completion_stream():
    """Demo processing a chat completion stream using the raw gRPC API."""

    channel = grpc.secure_channel(DEFAULT_BASE_URL, grpc.ssl_channel_credentials())
    stub = apex_pb2_grpc.ApexServiceStub(channel)

    sampling_params = apex_pb2.SamplingParameters(
        temperature=0.7,
        top_p=0.9,
        max_new_tokens=256,
        do_sample=True,
    )

    messages = [
        apex_pb2.ChatMessage(
            role="user",
            content="Write a short story about a cosmonaut learning to paint.",
        )
    ]

    request = apex_pb2.ChatCompletionRequest(
        messages=messages,
        sampling_parameters=sampling_params,
        stream=True,
    )

    api_key = os.environ.get("APEX_API_KEY", os.environ.get("MACROCOSMOS_API_KEY"))
    metadata = [
        ("x-client-id", __package_name__),
        ("x-client-version", __version__),
        ("authorization", f"Bearer {api_key}"),
    ]

    try:
        response_stream = stub.ChatCompletionStream(
            request,
            metadata=metadata,
            timeout=30.0,
        )

        print("Streaming response:")
        full_content = ""

        for chunk in response_stream:
            if chunk.choices and chunk.choices[0].delta.content:
                content = chunk.choices[0].delta.content
                full_content += content
                print(content, end="", flush=True)

    except grpc.RpcError as e:
        print(f"RPC error: {e.code()}: {e.details()}")


if __name__ == "__main__":
    print("Testing non-streaming ChatCompletion:")
    demo_chat_completion()

    print("\n" + "=" * 50 + "\n")

    print("Testing streaming ChatCompletion:")
    demo_chat_completion_stream()



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class GetUsageRequest(_message.Message):
    __slots__ = ("product_type",)
    PRODUCT_TYPE_FIELD_NUMBER: _ClassVar[int]
    product_type: str
    def __init__(self, product_type: _Optional[str] = ...) -> None: ...

class BillingRate(_message.Message):
    __slots__ = ("rate_type", "unit_size", "unit_type", "price_per_unit", "currency")
    RATE_TYPE_FIELD_NUMBER: _ClassVar[int]
    UNIT_SIZE_FIELD_NUMBER: _ClassVar[int]
    UNIT_TYPE_FIELD_NUMBER: _ClassVar[int]
    PRICE_PER_UNIT_FIELD_NUMBER: _ClassVar[int]
    CURRENCY_FIELD_NUMBER: _ClassVar[int]
    rate_type: str
    unit_size: int
    unit_type: str
    price_per_unit: float
    currency: str
    def __init__(self, rate_type: _Optional[str] = ..., unit_size: _Optional[int] = ..., unit_type: _Optional[str] = ..., price_per_unit: _Optional[float] = ..., currency: _Optional[str] = ...) -> None: ...

class GetUsageResponse(_message.Message):
    __slots__ = ("available_credits", "used_credits", "remaining_credits", "billing_rates")
    AVAILABLE_CREDITS_FIELD_NUMBER: _ClassVar[int]
    USED_CREDITS_FIELD_NUMBER: _ClassVar[int]
    REMAINING_CREDITS_FIELD_NUMBER: _ClassVar[int]
    BILLING_RATES_FIELD_NUMBER: _ClassVar[int]
    available_credits: float
    used_credits: float
    remaining_credits: float
    billing_rates: _containers.RepeatedCompositeFieldContainer[BillingRate]
    def __init__(self, available_credits: _Optional[float] = ..., used_credits: _Optional[float] = ..., remaining_credits: _Optional[float] = ..., billing_rates: _Optional[_Iterable[_Union[BillingRate, _Mapping]]] = ...) -> None: ...



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings
from macrocosmos.generated.apex.v1 import apex_pb2 as apex_dot_v1_dot_apex__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2

GRPC_GENERATED_VERSION = '1.71.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in apex/v1/apex_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class ApexServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ChatCompletion = channel.unary_unary(
                '/apex.v1.ApexService/ChatCompletion',
                request_serializer=apex_dot_v1_dot_apex__pb2.ChatCompletionRequest.SerializeToString,
                response_deserializer=apex_dot_v1_dot_apex__pb2.ChatCompletionResponse.FromString,
                _registered_method=True)
        self.ChatCompletionStream = channel.unary_stream(
                '/apex.v1.ApexService/ChatCompletionStream',
                request_serializer=apex_dot_v1_dot_apex__pb2.ChatCompletionRequest.SerializeToString,
                response_deserializer=apex_dot_v1_dot_apex__pb2.ChatCompletionChunkResponse.FromString,
                _registered_method=True)
        self.WebRetrieval = channel.unary_unary(
                '/apex.v1.ApexService/WebRetrieval',
                request_serializer=apex_dot_v1_dot_apex__pb2.WebRetrievalRequest.SerializeToString,
                response_deserializer=apex_dot_v1_dot_apex__pb2.WebRetrievalResponse.FromString,
                _registered_method=True)
        self.SubmitDeepResearcherJob = channel.unary_unary(
                '/apex.v1.ApexService/SubmitDeepResearcherJob',
                request_serializer=apex_dot_v1_dot_apex__pb2.ChatCompletionRequest.SerializeToString,
                response_deserializer=apex_dot_v1_dot_apex__pb2.SubmitDeepResearcherJobResponse.FromString,
                _registered_method=True)
        self.GetDeepResearcherJob = channel.unary_unary(
                '/apex.v1.ApexService/GetDeepResearcherJob',
                request_serializer=apex_dot_v1_dot_apex__pb2.GetDeepResearcherJobRequest.SerializeToString,
                response_deserializer=apex_dot_v1_dot_apex__pb2.GetDeepResearcherJobResponse.FromString,
                _registered_method=True)
        self.GetChatSessions = channel.unary_unary(
                '/apex.v1.ApexService/GetChatSessions',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=apex_dot_v1_dot_apex__pb2.GetChatSessionsResponse.FromString,
                _registered_method=True)
        self.UpdateChatAttribute = channel.unary_unary(
                '/apex.v1.ApexService/UpdateChatAttribute',
                request_serializer=apex_dot_v1_dot_apex__pb2.UpdateChatAttributeRequest.SerializeToString,
                response_deserializer=apex_dot_v1_dot_apex__pb2.UpdateChatAttributeResponse.FromString,
                _registered_method=True)


class ApexServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def ChatCompletion(self, request, context):
        """ChatCompletion generates a completion for a given request.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ChatCompletionStream(self, request, context):
        """ChatCompletionStream generates a stream of completions for a given request.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def WebRetrieval(self, request, context):
        """WebRetrieval retrieves web search results for a given request.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SubmitDeepResearcherJob(self, request, context):
        """SubmitDeepResearcherJob submits a new deep researcher job for processing.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetDeepResearcherJob(self, request, context):
        """GetDeepResearcherJob retrieves the status and results of a deep researcher job.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetChatSessions(self, request, context):
        """GetChatSessions retrieves a user's chats
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateChatAttribute(self, request, context):
        """UpdateChatAttribute updates attribute after asking LLM for one
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ApexServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ChatCompletion': grpc.unary_unary_rpc_method_handler(
                    servicer.ChatCompletion,
                    request_deserializer=apex_dot_v1_dot_apex__pb2.ChatCompletionRequest.FromString,
                    response_serializer=apex_dot_v1_dot_apex__pb2.ChatCompletionResponse.SerializeToString,
            ),
            'ChatCompletionStream': grpc.unary_stream_rpc_method_handler(
                    servicer.ChatCompletionStream,
                    request_deserializer=apex_dot_v1_dot_apex__pb2.ChatCompletionRequest.FromString,
                    response_serializer=apex_dot_v1_dot_apex__pb2.ChatCompletionChunkResponse.SerializeToString,
            ),
            'WebRetrieval': grpc.unary_unary_rpc_method_handler(
                    servicer.WebRetrieval,
                    request_deserializer=apex_dot_v1_dot_apex__pb2.WebRetrievalRequest.FromString,
                    response_serializer=apex_dot_v1_dot_apex__pb2.WebRetrievalResponse.SerializeToString,
            ),
            'SubmitDeepResearcherJob': grpc.unary_unary_rpc_method_handler(
                    servicer.SubmitDeepResearcherJob,
                    request_deserializer=apex_dot_v1_dot_apex__pb2.ChatCompletionRequest.FromString,
                    response_serializer=apex_dot_v1_dot_apex__pb2.SubmitDeepResearcherJobResponse.SerializeToString,
            ),
            'GetDeepResearcherJob': grpc.unary_unary_rpc_method_handler(
                    servicer.GetDeepResearcherJob,
                    request_deserializer=apex_dot_v1_dot_apex__pb2.GetDeepResearcherJobRequest.FromString,
                    response_serializer=apex_dot_v1_dot_apex__pb2.GetDeepResearcherJobResponse.SerializeToString,
            ),
            'GetChatSessions': grpc.unary_unary_rpc_method_handler(
                    servicer.GetChatSessions,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=apex_dot_v1_dot_apex__pb2.GetChatSessionsResponse.SerializeToString,
            ),
            'UpdateChatAttribute': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateChatAttribute,
                    request_deserializer=apex_dot_v1_dot_apex__pb2.UpdateChatAttributeRequest.FromString,
                    response_serializer=apex_dot_v1_dot_apex__pb2.UpdateChatAttributeResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'apex.v1.ApexService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('apex.v1.ApexService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class ApexService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def ChatCompletion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/apex.v1.ApexService/ChatCompletion',
            apex_dot_v1_dot_apex__pb2.ChatCompletionRequest.SerializeToString,
            apex_dot_v1_dot_apex__pb2.ChatCompletionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ChatCompletionStream(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/apex.v1.ApexService/ChatCompletionStream',
            apex_dot_v1_dot_apex__pb2.ChatCompletionRequest.SerializeToString,
            apex_dot_v1_dot_apex__pb2.ChatCompletionChunkResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def WebRetrieval(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/apex.v1.ApexService/WebRetrieval',
            apex_dot_v1_dot_apex__pb2.WebRetrievalRequest.SerializeToString,
            apex_dot_v1_dot_apex__pb2.WebRetrievalResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SubmitDeepResearcherJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/apex.v1.ApexService/SubmitDeepResearcherJob',
            apex_dot_v1_dot_apex__pb2.ChatCompletionRequest.SerializeToString,
            apex_dot_v1_dot_apex__pb2.SubmitDeepResearcherJobResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetDeepResearcherJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/apex.v1.ApexService/GetDeepResearcherJob',
            apex_dot_v1_dot_apex__pb2.GetDeepResearcherJobRequest.SerializeToString,
            apex_dot_v1_dot_apex__pb2.GetDeepResearcherJobResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetChatSessions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/apex.v1.ApexService/GetChatSessions',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            apex_dot_v1_dot_apex__pb2.GetChatSessionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UpdateChatAttribute(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/apex.v1.ApexService/UpdateChatAttribute',
            apex_dot_v1_dot_apex__pb2.UpdateChatAttributeRequest.SerializeToString,
            apex_dot_v1_dot_apex__pb2.UpdateChatAttributeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"""
Example of using the SN13 On Demand Data Streaming service with Macrocosmos SDK.
"""

import os

import macrocosmos as mc

api_key = os.environ.get("SN13_API_KEY", os.environ.get("MACROCOSMOS_API_KEY"))

client = mc.Sn13Client(api_key=api_key, app_name="examples/sn13_on_demand_data.py")

response = client.sn13.OnDemandData(
    source="x",
    usernames=["nasa", "spacex"],
    keywords=["photo", "space", "mars"],
    start_date="2024-04-01",
    end_date="2025-04-25",
    limit=3,
)

print(response)



import macrocosmos as mc
import os


def main():
    # Initialize the client
    client = mc.BillingClient(api_key=os.environ.get("MACROCOSMOS_API_KEY"))

    # Get usage information for gravity
    response = client.billing.GetUsage()

    # Print the response
    print(f"Available credits: {response.available_credits}")
    print(f"Used credits: {response.used_credits}")
    print(f"Remaining credits: {response.remaining_credits}")
    print("\nBilling rates:")
    for rate in response.billing_rates:
        print(f"  - Rate type: {rate.rate_type}")
        print(f"    Unit size: {rate.unit_size}")
        print(f"    Unit type: {rate.unit_type}")
        print(f"    Price per unit: {rate.price_per_unit} {rate.currency}")


if __name__ == "__main__":
    main()



# This is an automatically generated file, please do not change
# gen by protobuf_to_pydantic[v0.3.1.1](https://github.com/so1n/protobuf_to_pydantic)
# Protobuf Version: 5.29.4 
# Pydantic Version: 2.11.0 
from datetime import datetime
from google.protobuf.message import Message  # type: ignore
from pydantic import BaseModel
from pydantic import Field
import typing


class ChatMessage(BaseModel):
    """
     The choice object for the chunk.
 Parsed from https://github.com/macrocosm-os/prompting/blob/main/validator_api/serializers.py
    """

# role: the role of the message.
    role: str = Field(default="")
# content: the content of the message.
    content: str = Field(default="")

class SamplingParameters(BaseModel):
    """
     The sampling parameters for the completion.
 Parsed from https://github.com/macrocosm-os/prompting/blob/main/validator_api/serializers.py
    """

# temperature: the temperature to use for the completion.
    temperature: float = Field(default=0.0)
# top_p: the top_p to use for the completion.
    top_p: float = Field(default=0.0)
# top_k: the top_k to use for the completion.
    top_k: typing.Optional[float] = Field(default=0.0)
# max_new_tokens: the max_new_tokens to use for the completion.
    max_new_tokens: int = Field(default=0)
# do_sample: whether to do sample for the completion.
    do_sample: bool = Field(default=False)

class ChatCompletionRequest(BaseModel):
    """
     A request to generate completions following Apex CompletionsRequest format.
 Parsed from https://github.com/macrocosm-os/prompting/blob/main/validator_api/serializers.py
    """

# uids: the miner UIDs that will be used to generate the completion (optional).
    uids: typing.List[int] = Field(default_factory=list)
# messages: the messages to generate completions for.
    messages: typing.List[ChatMessage] = Field(default_factory=list)
# seed: the seed to use for the completion.
    seed: typing.Optional[int] = Field(default=0)
# task: the task to generate completions for (e.g. "InferenceTask").
    task: typing.Optional[str] = Field(default="")
# model: the LLM name to use for the completion. (optional, suggest leaving this empty as not all LLMs are supported)
    model: typing.Optional[str] = Field(default="")
# test_time_inference: whether to use test time inference.
    test_time_inference: typing.Optional[bool] = Field(default=False)
# mixture: whether to use a mixture of miners to create a slower but better answer.
    mixture: typing.Optional[bool] = Field(default=False)
# sampling_parameters: the sampling parameters to use for the completion.
    sampling_parameters: typing.Optional[SamplingParameters] = Field(default_factory=SamplingParameters)
# inference_mode: the inference mode to use for the completion.
    inference_mode: typing.Optional[str] = Field(default="")
# json_format: whether to use JSON format for the completion.
    json_format: typing.Optional[bool] = Field(default=False)
# stream: whether to stream the completion.
    stream: typing.Optional[bool] = Field(default=False)
# timeout: the timeout for the completion in seconds.
    timeout: typing.Optional[int] = Field(default=0)

class TopLogprob(BaseModel):
    """
     The top logprob object for the token. (not currently supported in Apex)
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_token_logprob.py
    """

# token: the token of the logprob.
    token: str = Field(default="")
# bytes: the bytes of the logprob.
    bytes: typing.List[int] = Field(default_factory=list)
# logprob: the logprob of the token.
    logprob: float = Field(default=0.0)

class ChatCompletionTokenLogprob(BaseModel):
    """
     The chat completion token logprob object. (not currently supported in Apex)
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_token_logprob.py
    """

# token: the token of the logprob.
    token: str = Field(default="")
# bytes: the bytes of the logprob.
    bytes: typing.List[int] = Field(default_factory=list)
# logprob: the logprob of the token.
    logprob: float = Field(default=0.0)
# top_logprobs: the top logprobs of the token.
    top_logprobs: typing.List[TopLogprob] = Field(default_factory=list)

class ChoiceLogprobs(BaseModel):
    """
     The logprobs object for the choice.
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_chunk.py
    """

# content: the content of the logprobs.
    content: typing.List[ChatCompletionTokenLogprob] = Field(default_factory=list)
# refusal: the refusal of the logprobs.
    refusal: typing.List[ChatCompletionTokenLogprob] = Field(default_factory=list)

class Annotation(BaseModel):
    """
     The annotation object for the message. (not currently supported in Apex)
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_message.py
    """

# content: the content of the annotation.
    content: str = Field(default="")
# role: the role of the annotation.
    role: str = Field(default="")

class ChatCompletionAudio(BaseModel):
    """
     The audio object for the message. (not currently supported in Apex)
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_audio.py
    """

# id: the id of the audio.
    id: str = Field(default="")
# data: the data of the audio.
    data: str = Field(default="")
# expires_at: the expires at of the audio.
    expires_at: int = Field(default=0)
# transcript: the transcript of the audio.
    transcript: str = Field(default="")

class FunctionCall(BaseModel):
    """
     The function call object for the message.
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_message.py
    """

# arguments: the arguments of the function call.
    arguments: typing.List[str] = Field(default_factory=list)
# name: the name of the function call.
    name: str = Field(default="")

class Function(BaseModel):
    """
     The function object for the tool call.
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_message_tool_call.py
    """

# arguments: the arguments of the function.
    arguments: typing.List[str] = Field(default_factory=list)
# name: the name of the function.
    name: str = Field(default="")

class ChatCompletionMessageToolCall(BaseModel):
    """
     The tool call object for the message.
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_message_tool_call.py
    """

# id: the id of the tool call.
    id: str = Field(default="")
# function: the function object for the tool call.
    function: Function = Field(default_factory=Function)
# type: the type of the tool call.
    type: str = Field(default="")

class ChatCompletionMessage(BaseModel):
    """
     The message response object from the LLM.
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_message.py
    """

# content: the content of the message.
    content: str = Field(default="")
# refusal: the refusal of the message. (not currently supported in Apex)
    refusal: str = Field(default="")
# role: the role of the message.
    role: str = Field(default="")
# annotations: the annotations of the message. (not currently supported in Apex)
    annotations: typing.List[Annotation] = Field(default_factory=list)
# audio: the audio of the message. (not currently supported in Apex)
    audio: ChatCompletionAudio = Field(default_factory=ChatCompletionAudio)
# function_call: the function call of the message.
    function_call: FunctionCall = Field(default_factory=FunctionCall)
# tool_calls: the tool calls of the message.
    tool_calls: typing.List[ChatCompletionMessageToolCall] = Field(default_factory=list)

class Choice(BaseModel):
    """
     The choice object containing the message response from the LLM for the completion.
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion.py
    """

# finish_reason: the finish reason of the choice.
    finish_reason: str = Field(default="")
# index: the index of the choice.
    index: int = Field(default=0)
# logprobs: the logprobs of the choice.
    logprobs: ChoiceLogprobs = Field(default_factory=ChoiceLogprobs)
# message: the message of the choice.
    message: ChatCompletionMessage = Field(default_factory=ChatCompletionMessage)

class CompletionTokensDetails(BaseModel):
    """
     The completion tokens details object. (not currently supported in Apex)
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/completion_usage.py
    """

# accepted_prediction_tokens: the accepted prediction tokens of the details.
    accepted_prediction_tokens: int = Field(default=0)
# audio_tokens: the audio tokens of the details.
    audio_tokens: int = Field(default=0)
# reasoning_tokens: the reasoning tokens of the details.
    reasoning_tokens: int = Field(default=0)
# rejected_prediction_tokens: the rejected prediction tokens of the details.
    rejected_prediction_tokens: int = Field(default=0)

class PromptTokensDetails(BaseModel):
    """
     The prompt tokens details object. (not currently supported in Apex)
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/completion_usage.py
    """

# audio_tokens: the audio tokens of the details.
    audio_tokens: int = Field(default=0)
# cached_tokens: the cached tokens of the details.
    cached_tokens: int = Field(default=0)

class CompletionUsage(BaseModel):
    """
     The completion usage object. (not currently supported in Apex)
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/completion_usage.py
    """

# completion_tokens: the completion tokens of the usage.
    completion_tokens: int = Field(default=0)
# prompt_tokens: the prompt tokens of the usage.
    prompt_tokens: int = Field(default=0)
# total_tokens: the total tokens of the usage.
    total_tokens: int = Field(default=0)
# completion_tokens_details: the completion tokens details of the usage.
    completion_tokens_details: CompletionTokensDetails = Field(default_factory=CompletionTokensDetails)
# prompt_tokens_details: the prompt tokens details of the usage.
    prompt_tokens_details: PromptTokensDetails = Field(default_factory=PromptTokensDetails)

class ChatCompletionResponse(BaseModel):
    """
     A chat completion response, following OpenAI's ChatCompletion format.
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion.py
    """

# id: the id of the completion.
    id: str = Field(default="")
# choices: the choices of the completion.
    choices: typing.List[Choice] = Field(default_factory=list)
# created: the created time of the completion.
    created: int = Field(default=0)
# model: the model of the completion.
    model: str = Field(default="")
# object: the object of the completion.
    object: str = Field(default="")
# service_tier: the service tier of the completion. (not currently supported in Apex)
    service_tier: str = Field(default="")
# system_fingerprint: the system fingerprint of the completion. (not currently supported in Apex)
    system_fingerprint: str = Field(default="")
# usage: the usage of the completion. (not currently supported in Apex)
    usage: CompletionUsage = Field(default_factory=CompletionUsage)

class ChoiceDeltaFunctionCall(BaseModel):
    """
     The function call object for the delta.
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_chunk.py
    """

# arguments: the arguments of the function call.
    arguments: typing.List[str] = Field(default_factory=list)
# name: the name of the function call.
    name: str = Field(default="")

class ChoiceDeltaToolCallFunction(BaseModel):
    """
     The function object for the tool call.
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_chunk.py
    """

# arguments: the arguments of the function.
    arguments: typing.List[str] = Field(default_factory=list)
# name: the name of the function.
    name: str = Field(default="")

class ChoiceDeltaToolCall(BaseModel):
    """
     The tool call object for the delta.
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_chunk.py
    """

# index: the index of the tool call.
    index: int = Field(default=0)
# id: the id of the tool call.
    id: str = Field(default="")
# function: the function object for the tool call.
    function: ChoiceDeltaToolCallFunction = Field(default_factory=ChoiceDeltaToolCallFunction)
# type: the type of the tool call.
    type: str = Field(default="")

class ChoiceDelta(BaseModel):
    """
     The delta object for the choice.
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_chunk.py
    """

# content: the content of the delta.
    content: str = Field(default="")
# function_call: the function call of the delta.
    function_call: ChoiceDeltaFunctionCall = Field(default_factory=ChoiceDeltaFunctionCall)
# refusal: the refusal of the delta.
    refusal: str = Field(default="")
# role: the role of the delta.
    role: str = Field(default="")
# tool_calls: the tool calls of the delta.
    tool_calls: typing.List[ChoiceDeltaToolCall] = Field(default_factory=list)

class ChunkChoice(BaseModel):
    """
     The choice object for the chunk.
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_chunk.py
    """

# delta: the delta of the choice.
    delta: ChoiceDelta = Field(default_factory=ChoiceDelta)
# finish_reason: the finish reason of the choice.
    finish_reason: str = Field(default="")
# index: the index of the choice.
    index: int = Field(default=0)
# logprobs: the logprobs of the choice. (not currently supported in Apex)
    logprobs: ChoiceLogprobs = Field(default_factory=ChoiceLogprobs)

class ChatCompletionChunkResponse(BaseModel):
    """
     A streaming chunk response, following OpenAI's ChatCompletionChunk format.
 Parsed from https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_chunk.py
    """

# id: the id of the chunk.
    id: str = Field(default="")
# choices: the choices of the chunk.
    choices: typing.List[ChunkChoice] = Field(default_factory=list)
# created: the created time of the chunk.
    created: int = Field(default=0)
# model: the model of the chunk.
    model: str = Field(default="")
# object: the object of the chunk. (not currently supported in Apex)
    object: str = Field(default="")
# service_tier: the service tier of the chunk. (not currently supported in Apex)
    service_tier: str = Field(default="")
# system_fingerprint: the system fingerprint of the chunk. (not currently supported in Apex)
    system_fingerprint: str = Field(default="")
# usage: the usage of the chunk. (not currently supported in Apex)
    usage: CompletionUsage = Field(default_factory=CompletionUsage)

class WebRetrievalRequest(BaseModel):
    """
     A web retrieval request from Apex
 Parsed from https://github.com/macrocosm-os/prompting/blob/main/validator_api/serializers.py
    """

# uids: the miner UIDs that will be used to generate the completion (optional).
    uids: typing.List[int] = Field(default_factory=list)
# search_query: the search query.
    search_query: str = Field(default="")
# n_miners: the number of miners to use for the query.
    n_miners: typing.Optional[int] = Field(default=0)
# n_results: the number of results to return.
    n_results: typing.Optional[int] = Field(default=0)
# max_response_time: the max response time to allow for the miners to respond in seconds.
    max_response_time: typing.Optional[int] = Field(default=0)
# timeout: the timeout for the web retrieval in seconds.
    timeout: typing.Optional[int] = Field(default=0)

class WebSearchResult(BaseModel):
    """
     A web search result from Apex
 Parsed from https://github.com/macrocosm-os/prompting/blob/main/validator_api/serializers.py
    """

# url: the url of the result.
    url: str = Field(default="")
# content: the entire page contents.
    content: str = Field(default="")
# relevant: the relevant part of the page best fitting the query..
    relevant: str = Field(default="")

class WebRetrievalResponse(BaseModel):
    """
     A web retrieval response from Apex
 Parsed from https://github.com/macrocosm-os/prompting/blob/main/validator_api/serializers.py
    """

# results: the results of the web retrieval.
    results: typing.List[WebSearchResult] = Field(default_factory=list)

class SubmitDeepResearcherJobResponse(BaseModel):
    """
     A response containing the deep researcher job submission details
    """

# job_id: unique identifier for the submitted job
    job_id: str = Field(default="")
# status: current status of the job
    status: str = Field(default="")
# created_at: timestamp when the job was created
    created_at: str = Field(default="")
# updated_at: timestamp when the job was last updated
    updated_at: str = Field(default="")

class DeepResearcherResultChunk(BaseModel):
    """
     A chunk of the deep researcher result
    """

# seq_id: sequence identifier for the chunk
    seq_id: int = Field(default=0)
# chunk: the content of the chunk
    chunk: str = Field(default="")

class GetDeepResearcherJobResponse(BaseModel):
    """
     A response containing the deep researcher job status and results
    """

# job_id: unique identifier for the job
    job_id: str = Field(default="")
# status: current status of the job
    status: str = Field(default="")
# created_at: timestamp when the job was created
    created_at: str = Field(default="")
# updated_at: timestamp when the job was last updated
    updated_at: str = Field(default="")
# result: array of result chunks
    result: typing.List[DeepResearcherResultChunk] = Field(default_factory=list)
# error: error message if the job failed
    error: typing.Optional[str] = Field(default="")

class GetDeepResearcherJobRequest(BaseModel):
    """
     A request to get the status of a deep researcher job
    """

# job_id: the ID of the job to retrieve
    job_id: str = Field(default="")

class ChatSession(BaseModel):
    """
     A GetChatSession message repeated in GetChatSessionsResponse
    """

# id: chat id
    id: str = Field(default="")
# user_id: user id
    user_id: str = Field(default="")
# title: title of chat
    title: str = Field(default="")
# chat_type: e.g. apex
    chat_type: str = Field(default="")
# created_at: when the chat was created
    created_at: datetime = Field(default_factory=datetime.now)
# updated_at: when the chat was updated
    updated_at: datetime = Field(default_factory=datetime.now)

class GetChatSessionsResponse(BaseModel):
    """
     A GetChatSessionsResponse response
    """

# chat_sessions: the chat sessions
    chat_sessions: typing.List[ChatSession] = Field(default_factory=list)

class UpdateChatAttributeRequest(BaseModel):
    """
     Directly model the attributes as a map
    """

# chat_id: the unique id associated to a users chat message
    chat_id: str = Field(default="")
# attributes: the data attributes captured in the chat logging process
    attributes: typing.Dict[str, str] = Field(default_factory=dict)

class UpdateChatAttributeResponse(BaseModel):
# success: indicates if an attribute update was successful
    success: bool = Field(default=False)



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Crawler(_message.Message):
    __slots__ = ("crawler_id", "criteria", "start_time", "deregistration_time", "archive_time", "state", "dataset_workflows")
    CRAWLER_ID_FIELD_NUMBER: _ClassVar[int]
    CRITERIA_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    DEREGISTRATION_TIME_FIELD_NUMBER: _ClassVar[int]
    ARCHIVE_TIME_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    DATASET_WORKFLOWS_FIELD_NUMBER: _ClassVar[int]
    crawler_id: str
    criteria: CrawlerCriteria
    start_time: _timestamp_pb2.Timestamp
    deregistration_time: _timestamp_pb2.Timestamp
    archive_time: _timestamp_pb2.Timestamp
    state: CrawlerState
    dataset_workflows: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, crawler_id: _Optional[str] = ..., criteria: _Optional[_Union[CrawlerCriteria, _Mapping]] = ..., start_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., deregistration_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., archive_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., state: _Optional[_Union[CrawlerState, _Mapping]] = ..., dataset_workflows: _Optional[_Iterable[str]] = ...) -> None: ...

class CrawlerCriteria(_message.Message):
    __slots__ = ("platform", "topic", "notification", "mock")
    PLATFORM_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    NOTIFICATION_FIELD_NUMBER: _ClassVar[int]
    MOCK_FIELD_NUMBER: _ClassVar[int]
    platform: str
    topic: str
    notification: CrawlerNotification
    mock: bool
    def __init__(self, platform: _Optional[str] = ..., topic: _Optional[str] = ..., notification: _Optional[_Union[CrawlerNotification, _Mapping]] = ..., mock: bool = ...) -> None: ...

class CrawlerNotification(_message.Message):
    __slots__ = ("to", "link")
    TO_FIELD_NUMBER: _ClassVar[int]
    LINK_FIELD_NUMBER: _ClassVar[int]
    to: str
    link: str
    def __init__(self, to: _Optional[str] = ..., link: _Optional[str] = ...) -> None: ...

class HfRepo(_message.Message):
    __slots__ = ("repo_name", "row_count", "last_update")
    REPO_NAME_FIELD_NUMBER: _ClassVar[int]
    ROW_COUNT_FIELD_NUMBER: _ClassVar[int]
    LAST_UPDATE_FIELD_NUMBER: _ClassVar[int]
    repo_name: str
    row_count: int
    last_update: str
    def __init__(self, repo_name: _Optional[str] = ..., row_count: _Optional[int] = ..., last_update: _Optional[str] = ...) -> None: ...

class CrawlerState(_message.Message):
    __slots__ = ("status", "bytes_collected", "records_collected", "repos")
    STATUS_FIELD_NUMBER: _ClassVar[int]
    BYTES_COLLECTED_FIELD_NUMBER: _ClassVar[int]
    RECORDS_COLLECTED_FIELD_NUMBER: _ClassVar[int]
    REPOS_FIELD_NUMBER: _ClassVar[int]
    status: str
    bytes_collected: int
    records_collected: int
    repos: _containers.RepeatedCompositeFieldContainer[HfRepo]
    def __init__(self, status: _Optional[str] = ..., bytes_collected: _Optional[int] = ..., records_collected: _Optional[int] = ..., repos: _Optional[_Iterable[_Union[HfRepo, _Mapping]]] = ...) -> None: ...

class GravityTaskState(_message.Message):
    __slots__ = ("gravity_task_id", "name", "status", "start_time", "crawler_ids", "crawler_workflows")
    GRAVITY_TASK_ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    CRAWLER_IDS_FIELD_NUMBER: _ClassVar[int]
    CRAWLER_WORKFLOWS_FIELD_NUMBER: _ClassVar[int]
    gravity_task_id: str
    name: str
    status: str
    start_time: _timestamp_pb2.Timestamp
    crawler_ids: _containers.RepeatedScalarFieldContainer[str]
    crawler_workflows: _containers.RepeatedCompositeFieldContainer[Crawler]
    def __init__(self, gravity_task_id: _Optional[str] = ..., name: _Optional[str] = ..., status: _Optional[str] = ..., start_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., crawler_ids: _Optional[_Iterable[str]] = ..., crawler_workflows: _Optional[_Iterable[_Union[Crawler, _Mapping]]] = ...) -> None: ...

class GetGravityTasksRequest(_message.Message):
    __slots__ = ("gravity_task_id", "include_crawlers")
    GRAVITY_TASK_ID_FIELD_NUMBER: _ClassVar[int]
    INCLUDE_CRAWLERS_FIELD_NUMBER: _ClassVar[int]
    gravity_task_id: str
    include_crawlers: bool
    def __init__(self, gravity_task_id: _Optional[str] = ..., include_crawlers: bool = ...) -> None: ...

class GetGravityTasksResponse(_message.Message):
    __slots__ = ("gravity_task_states",)
    GRAVITY_TASK_STATES_FIELD_NUMBER: _ClassVar[int]
    gravity_task_states: _containers.RepeatedCompositeFieldContainer[GravityTaskState]
    def __init__(self, gravity_task_states: _Optional[_Iterable[_Union[GravityTaskState, _Mapping]]] = ...) -> None: ...

class GravityTask(_message.Message):
    __slots__ = ("topic", "platform")
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    PLATFORM_FIELD_NUMBER: _ClassVar[int]
    topic: str
    platform: str
    def __init__(self, topic: _Optional[str] = ..., platform: _Optional[str] = ...) -> None: ...

class NotificationRequest(_message.Message):
    __slots__ = ("type", "address", "redirect_url")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    REDIRECT_URL_FIELD_NUMBER: _ClassVar[int]
    type: str
    address: str
    redirect_url: str
    def __init__(self, type: _Optional[str] = ..., address: _Optional[str] = ..., redirect_url: _Optional[str] = ...) -> None: ...

class GetCrawlerRequest(_message.Message):
    __slots__ = ("crawler_id",)
    CRAWLER_ID_FIELD_NUMBER: _ClassVar[int]
    crawler_id: str
    def __init__(self, crawler_id: _Optional[str] = ...) -> None: ...

class GetCrawlerResponse(_message.Message):
    __slots__ = ("crawler",)
    CRAWLER_FIELD_NUMBER: _ClassVar[int]
    crawler: Crawler
    def __init__(self, crawler: _Optional[_Union[Crawler, _Mapping]] = ...) -> None: ...

class CreateGravityTaskRequest(_message.Message):
    __slots__ = ("gravity_tasks", "name", "notification_requests", "gravity_task_id")
    GRAVITY_TASKS_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    NOTIFICATION_REQUESTS_FIELD_NUMBER: _ClassVar[int]
    GRAVITY_TASK_ID_FIELD_NUMBER: _ClassVar[int]
    gravity_tasks: _containers.RepeatedCompositeFieldContainer[GravityTask]
    name: str
    notification_requests: _containers.RepeatedCompositeFieldContainer[NotificationRequest]
    gravity_task_id: str
    def __init__(self, gravity_tasks: _Optional[_Iterable[_Union[GravityTask, _Mapping]]] = ..., name: _Optional[str] = ..., notification_requests: _Optional[_Iterable[_Union[NotificationRequest, _Mapping]]] = ..., gravity_task_id: _Optional[str] = ...) -> None: ...

class CreateGravityTaskResponse(_message.Message):
    __slots__ = ("gravity_task_id",)
    GRAVITY_TASK_ID_FIELD_NUMBER: _ClassVar[int]
    gravity_task_id: str
    def __init__(self, gravity_task_id: _Optional[str] = ...) -> None: ...

class BuildDatasetRequest(_message.Message):
    __slots__ = ("crawler_id", "notification_requests", "max_rows")
    CRAWLER_ID_FIELD_NUMBER: _ClassVar[int]
    NOTIFICATION_REQUESTS_FIELD_NUMBER: _ClassVar[int]
    MAX_ROWS_FIELD_NUMBER: _ClassVar[int]
    crawler_id: str
    notification_requests: _containers.RepeatedCompositeFieldContainer[NotificationRequest]
    max_rows: int
    def __init__(self, crawler_id: _Optional[str] = ..., notification_requests: _Optional[_Iterable[_Union[NotificationRequest, _Mapping]]] = ..., max_rows: _Optional[int] = ...) -> None: ...

class BuildDatasetResponse(_message.Message):
    __slots__ = ("dataset_id", "dataset")
    DATASET_ID_FIELD_NUMBER: _ClassVar[int]
    DATASET_FIELD_NUMBER: _ClassVar[int]
    dataset_id: str
    dataset: Dataset
    def __init__(self, dataset_id: _Optional[str] = ..., dataset: _Optional[_Union[Dataset, _Mapping]] = ...) -> None: ...

class Nebula(_message.Message):
    __slots__ = ("error", "file_size_bytes", "url")
    ERROR_FIELD_NUMBER: _ClassVar[int]
    FILE_SIZE_BYTES_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    error: str
    file_size_bytes: int
    url: str
    def __init__(self, error: _Optional[str] = ..., file_size_bytes: _Optional[int] = ..., url: _Optional[str] = ...) -> None: ...

class Dataset(_message.Message):
    __slots__ = ("crawler_workflow_id", "create_date", "expire_date", "files", "status", "status_message", "steps", "total_steps", "nebula")
    CRAWLER_WORKFLOW_ID_FIELD_NUMBER: _ClassVar[int]
    CREATE_DATE_FIELD_NUMBER: _ClassVar[int]
    EXPIRE_DATE_FIELD_NUMBER: _ClassVar[int]
    FILES_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    STATUS_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    STEPS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_STEPS_FIELD_NUMBER: _ClassVar[int]
    NEBULA_FIELD_NUMBER: _ClassVar[int]
    crawler_workflow_id: str
    create_date: _timestamp_pb2.Timestamp
    expire_date: _timestamp_pb2.Timestamp
    files: _containers.RepeatedCompositeFieldContainer[DatasetFile]
    status: str
    status_message: str
    steps: _containers.RepeatedCompositeFieldContainer[DatasetStep]
    total_steps: int
    nebula: Nebula
    def __init__(self, crawler_workflow_id: _Optional[str] = ..., create_date: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., expire_date: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., files: _Optional[_Iterable[_Union[DatasetFile, _Mapping]]] = ..., status: _Optional[str] = ..., status_message: _Optional[str] = ..., steps: _Optional[_Iterable[_Union[DatasetStep, _Mapping]]] = ..., total_steps: _Optional[int] = ..., nebula: _Optional[_Union[Nebula, _Mapping]] = ...) -> None: ...

class DatasetFile(_message.Message):
    __slots__ = ("file_name", "file_size_bytes", "last_modified", "num_rows", "s3_key", "url")
    FILE_NAME_FIELD_NUMBER: _ClassVar[int]
    FILE_SIZE_BYTES_FIELD_NUMBER: _ClassVar[int]
    LAST_MODIFIED_FIELD_NUMBER: _ClassVar[int]
    NUM_ROWS_FIELD_NUMBER: _ClassVar[int]
    S3_KEY_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    file_name: str
    file_size_bytes: int
    last_modified: _timestamp_pb2.Timestamp
    num_rows: int
    s3_key: str
    url: str
    def __init__(self, file_name: _Optional[str] = ..., file_size_bytes: _Optional[int] = ..., last_modified: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., num_rows: _Optional[int] = ..., s3_key: _Optional[str] = ..., url: _Optional[str] = ...) -> None: ...

class DatasetStep(_message.Message):
    __slots__ = ("progress", "step", "step_name")
    PROGRESS_FIELD_NUMBER: _ClassVar[int]
    STEP_FIELD_NUMBER: _ClassVar[int]
    STEP_NAME_FIELD_NUMBER: _ClassVar[int]
    progress: float
    step: int
    step_name: str
    def __init__(self, progress: _Optional[float] = ..., step: _Optional[int] = ..., step_name: _Optional[str] = ...) -> None: ...

class GetDatasetRequest(_message.Message):
    __slots__ = ("dataset_id",)
    DATASET_ID_FIELD_NUMBER: _ClassVar[int]
    dataset_id: str
    def __init__(self, dataset_id: _Optional[str] = ...) -> None: ...

class GetDatasetResponse(_message.Message):
    __slots__ = ("dataset",)
    DATASET_FIELD_NUMBER: _ClassVar[int]
    dataset: Dataset
    def __init__(self, dataset: _Optional[_Union[Dataset, _Mapping]] = ...) -> None: ...

class CancelGravityTaskRequest(_message.Message):
    __slots__ = ("gravity_task_id",)
    GRAVITY_TASK_ID_FIELD_NUMBER: _ClassVar[int]
    gravity_task_id: str
    def __init__(self, gravity_task_id: _Optional[str] = ...) -> None: ...

class CancelGravityTaskResponse(_message.Message):
    __slots__ = ("message",)
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class CancelDatasetRequest(_message.Message):
    __slots__ = ("dataset_id",)
    DATASET_ID_FIELD_NUMBER: _ClassVar[int]
    dataset_id: str
    def __init__(self, dataset_id: _Optional[str] = ...) -> None: ...

class CancelDatasetResponse(_message.Message):
    __slots__ = ("message",)
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class DatasetBillingCorrectionRequest(_message.Message):
    __slots__ = ("requested_row_count", "actual_row_count")
    REQUESTED_ROW_COUNT_FIELD_NUMBER: _ClassVar[int]
    ACTUAL_ROW_COUNT_FIELD_NUMBER: _ClassVar[int]
    requested_row_count: int
    actual_row_count: int
    def __init__(self, requested_row_count: _Optional[int] = ..., actual_row_count: _Optional[int] = ...) -> None: ...

class DatasetBillingCorrectionResponse(_message.Message):
    __slots__ = ("refund_amount",)
    REFUND_AMOUNT_FIELD_NUMBER: _ClassVar[int]
    refund_amount: float
    def __init__(self, refund_amount: _Optional[float] = ...) -> None: ...



"""
Example of using the Apex Web Search API with Macrocosmos SDK.
"""

import asyncio
import os

import macrocosmos as mc


# Synchronous example
def demo_web_search_sync():
    """Demo synchronous web search using the Macrocosmos SDK."""

    print("Running synchronous example")
    api_key = os.environ.get("APEX_API_KEY", os.environ.get("MACROCOSMOS_API_KEY"))
    client = mc.ApexClient(api_key=api_key, app_name="examples/apex_web_search")

    # Simple web search query
    response = client.web_search.search(
        search_query="What is Bittensor?",
        max_results_per_miner=3,
        max_response_time=20,
    )

    # Print the results
    print(f"Got {len(response.results)} results for 'What is Bittensor?'")
    for i, result in enumerate(response.results):
        print(f"\nResult {i + 1}:")
        print(f"URL: {result.url}")
        print(f"Relevant content: {result.relevant[:200]}...")


# Asynchronous example
async def demo_web_search_async():
    """Demo asynchronous web search using the Macrocosmos SDK."""

    print("\nRunning asynchronous example")
    api_key = os.environ.get("APEX_API_KEY", os.environ.get("MACROCOSMOS_API_KEY"))
    client = mc.AsyncApexClient(api_key=api_key, app_name="examples/apex_web_search")

    # Simple web search query
    response = await client.web_search.search(
        search_query="Latest AI research papers",
        max_results_per_miner=3,
        max_response_time=20,
    )

    # Print the results
    print(f"Got {len(response.results)} results for 'Latest AI research papers'")
    for i, result in enumerate(response.results):
        print(f"\nResult {i + 1}:")
        print(f"URL: {result.url}")
        print(f"Relevant content: {result.relevant[:200]}...")


if __name__ == "__main__":
    # Run synchronous example
    demo_web_search_sync()

    # Run asynchronous example using asyncio
    asyncio.run(demo_web_search_async())



# This is an automatically generated file, please do not change
# gen by protobuf_to_pydantic[v0.3.1.1](https://github.com/so1n/protobuf_to_pydantic)
# Protobuf Version: 5.29.4 
# Pydantic Version: 2.11.0 
from datetime import datetime
from google.protobuf.message import Message  # type: ignore
from pydantic import BaseModel
from pydantic import Field
import typing


class CrawlerNotification(BaseModel):
    """
     CrawlerNotification is the details of the notification to be sent to the user
    """

# to: the email address of the user
    to: str = Field(default="")
# link: the redirect link in the email where the user can view the dataset
    link: str = Field(default="")

class CrawlerCriteria(BaseModel):
    """
     CrawlerCriteria is the contents of the job and the notification details
    """

# platform: the platform of the job ('x' or 'reddit')
    platform: str = Field(default="")
# topic: the topic of the job (e.g. '#ai' for X, 'r/ai' for Reddit)
    topic: str = Field(default="")
# notification: the details of the notification to be sent to the user
    notification: CrawlerNotification = Field(default_factory=CrawlerNotification)
# mock: Used for testing purposes (optional, defaults to false)
    mock: bool = Field(default=False)

class HfRepo(BaseModel):
    """
     HfRepo is a single Hugging Face repository that contains data for a crawler
    """

# repo_name: the name of the Hugging Face repository
    repo_name: str = Field(default="")
# row_count: the number of rows in the repository for the crawler criteria
    row_count: int = Field(default=0)
# last_update: the last recorded time the repository was updated
    last_update: str = Field(default="")

class CrawlerState(BaseModel):
    """
     CrawlerState is the current state of the crawler
    """

# status: the current status of the crawler
#   "Pending"   -- Crawler is pending submission to the SN13 Validator
#   "Submitted" -- Crawler is submitted to the SN13 Validator
#   "Running"   -- Crawler is running (we got the first update)
#   "Completed" -- Crawler is completed (timer expired)
#   "Cancelled" -- Crawler is cancelled by user via cancellation of workflow
#   "Archived"  -- Crawler is archived (now read-only i.e. no new dataset)
#   "Failed"    -- Crawler failed to run
    status: str = Field(default="")
# bytes_collected: the estimated number of bytes collected by the crawler
    bytes_collected: int = Field(default=0)
# records_collected: the estimated number of records collected by the crawler
    records_collected: int = Field(default=0)
# repos: the Hugging Face repositories that contain data for a crawler
    repos: typing.List[HfRepo] = Field(default_factory=list)

class Crawler(BaseModel):
    """
     Crawler is a single crawler workflow that registers a single job (platform/topic) on SN13's dynamic desirability engine
    """

# crawler_id: the ID of the crawler
    crawler_id: str = Field(default="")
# criteria: the contents of the job and the notification details
    criteria: CrawlerCriteria = Field(default_factory=CrawlerCriteria)
# start_time: the time the crawler was created
    start_time: datetime = Field(default_factory=datetime.now)
# deregistration_time: the time the crawler was deregistered
    deregistration_time: datetime = Field(default_factory=datetime.now)
# archive_time: the time the crawler was archived
    archive_time: datetime = Field(default_factory=datetime.now)
# state: the current state of the crawler
    state: CrawlerState = Field(default_factory=CrawlerState)
# dataset_workflows: the IDs of the dataset workflows that are associated with the crawler
    dataset_workflows: typing.List[str] = Field(default_factory=list)

class GravityTaskState(BaseModel):
    """
     GravityTaskState is the current state of a gravity task
    """

# gravity_task_id: the ID of the gravity task
    gravity_task_id: str = Field(default="")
# name: the name given by the user of the gravity task
    name: str = Field(default="")
# status: the current status of the gravity task
    status: str = Field(default="")
# start_time: the time the gravity task was created
    start_time: datetime = Field(default_factory=datetime.now)
# crawler_ids: the IDs of the crawler workflows that are associated with the gravity task
    crawler_ids: typing.List[str] = Field(default_factory=list)
# crawler_workflows: the crawler workflows that are associated with the gravity task
    crawler_workflows: typing.List[Crawler] = Field(default_factory=list)

class GetGravityTasksRequest(BaseModel):
    """
     GetGravityTasksRequest is the request message for listing gravity tasks for a user
    """

# gravity_task_id: the ID of the gravity task (optional, if not provided, all gravity tasks for the user will be returned)
    gravity_task_id: typing.Optional[str] = Field(default="")
# include_crawlers: whether to include the crawler states in the response
    include_crawlers: typing.Optional[bool] = Field(default=False)

class GetGravityTasksResponse(BaseModel):
    """
     GetGravityTasksResponse is the response message for listing gravity tasks for a user
    """

# gravity_task_states: the current states of the gravity tasks
    gravity_task_states: typing.List[GravityTaskState] = Field(default_factory=list)

class GravityTask(BaseModel):
    """
     GravityTask defines a crawler's criteria for a single job (platform/topic)
    """

# topic: the topic of the job (e.g. '#ai' for X, 'r/ai' for Reddit)
    topic: str = Field(default="")
# platform: the platform of the job ('x' or 'reddit')
    platform: str = Field(default="")

class NotificationRequest(BaseModel):
    """
     NotificationRequest is the request message for sending a notification to a user when a dataset is ready to download
    """

# type: the type of notification to send ('email' is only supported currently)
    type: str = Field(default="")
# address: the address to send the notification to (only email addresses are supported currently)
    address: str = Field(default="")
# redirect_url: the URL to include in the notication message that redirects the user to any built datasets
    redirect_url: typing.Optional[str] = Field(default="")

class GetCrawlerRequest(BaseModel):
    """
     GetCrawlerRequest is the request message for getting a crawler
    """

# crawler_id: the ID of the crawler
    crawler_id: str = Field(default="")

class GetCrawlerResponse(BaseModel):
    """
     GetCrawlerResponse is the response message for getting a crawler
    """

# crawler: the crawler
    crawler: Crawler = Field(default_factory=Crawler)

class CreateGravityTaskRequest(BaseModel):
    """
     CreateGravityTaskRequest is the request message for creating a new gravity task
    """

# gravity_tasks: the criteria for the crawlers that will be created
    gravity_tasks: typing.List[GravityTask] = Field(default_factory=list)
# name: the name of the gravity task (optional, default will generate a random name)
    name: str = Field(default="")
# notification_requests: the details of the notification to be sent to the user when a dataset
#   that is automatically generated upon completion of the crawler is ready to download (optional)
    notification_requests: typing.List[NotificationRequest] = Field(default_factory=list)
# gravity_task_id: the ID of the gravity task (optional, default will generate a random ID)
    gravity_task_id: typing.Optional[str] = Field(default="")

class CreateGravityTaskResponse(BaseModel):
    """
     CreateGravityTaskResponse is the response message for creating a new gravity task
    """

# gravity_task_id: the ID of the gravity task
    gravity_task_id: str = Field(default="")

class BuildDatasetRequest(BaseModel):
    """
     BuildDatasetRequest is the request message for manually requesting the building of a dataset for a single crawler
    """

# crawler_id: the ID of the crawler that will be used to build the dataset
    crawler_id: str = Field(default="")
# notification_requests: the details of the notification to be sent to the user when the dataset is ready to download (optional)
    notification_requests: typing.List[NotificationRequest] = Field(default_factory=list)
# max_rows: the maximum number of rows to include in the dataset (optional, defaults to 500)
    max_rows: int = Field(default=0)

class DatasetFile(BaseModel):
    """
     DatasetFile contains the details about a dataset file
    """

# file_name: the name of the file
    file_name: str = Field(default="")
# file_size_bytes: the size of the file in bytes
    file_size_bytes: int = Field(default=0)
# last_modified: the date the file was last modified
    last_modified: datetime = Field(default_factory=datetime.now)
# num_rows: the number of rows in the file
    num_rows: int = Field(default=0)
# s3_key: the key of the file in S3 (internal use only)
    s3_key: str = Field(default="")
# url: the URL of the file (public use)
    url: str = Field(default="")

class DatasetStep(BaseModel):
    """
     DatasetStep contains one step of the progress of a dataset build
 (NOTE: each step varies in time and complexity)
    """

# progress: the progress of this step in the dataset build (0.0 - 1.0)
    progress: float = Field(default=0.0)
# step: the step number of the dataset build (1-indexed)
    step: int = Field(default=0)
# step_name: description of what is happening in the step
    step_name: str = Field(default="")

class Nebula(BaseModel):
# error: nebula build error message
    error: str = Field(default="")
# file_size_bytes: the size of the file in bytes
    file_size_bytes: int = Field(default=0)
# url: the URL of the file
    url: str = Field(default="")

class Dataset(BaseModel):
    """
     Dataset contains the progress and results of a dataset build
    """

# crawler_workflow_id: the ID of the parent crawler for this dataset
    crawler_workflow_id: str = Field(default="")
# create_date: the date the dataset was created
    create_date: datetime = Field(default_factory=datetime.now)
# expire_date: the date the dataset will expire (be deleted)
    expire_date: datetime = Field(default_factory=datetime.now)
# files: the details about the dataset files that are included in the dataset
    files: typing.List[DatasetFile] = Field(default_factory=list)
# status: the status of the dataset
    status: str = Field(default="")
# status_message: the message of the status of the dataset
    status_message: str = Field(default="")
# steps: the progress of the dataset build
    steps: typing.List[DatasetStep] = Field(default_factory=list)
# total_steps: the total number of steps in the dataset build
    total_steps: int = Field(default=0)
# nebula: the details about the nebula that was built
    nebula: Nebula = Field(default_factory=Nebula)

class BuildDatasetResponse(BaseModel):
    """
     BuildDatasetResponse is the response message for manually requesting the building of a dataset for a single crawler
 - dataset: the dataset that was built
    """

# dataset_id: the ID of the dataset
    dataset_id: str = Field(default="")
# dataset: the dataset that was built
    dataset: Dataset = Field(default_factory=Dataset)

class GetDatasetRequest(BaseModel):
    """
     GetDatasetRequest is the request message for getting the status of a dataset
    """

# dataset_id: the ID of the dataset
    dataset_id: str = Field(default="")

class GetDatasetResponse(BaseModel):
    """
     GetDatasetResponse is the response message for getting the status of a dataset
    """

# dataset: the dataset that is being built
    dataset: Dataset = Field(default_factory=Dataset)

class CancelGravityTaskRequest(BaseModel):
    """
     CancelGravityTaskRequest is the request message for cancelling a gravity task
    """

# gravity_task_id: the ID of the gravity task
    gravity_task_id: str = Field(default="")

class CancelGravityTaskResponse(BaseModel):
    """
     CancelGravityTaskResponse is the response message for cancelling a gravity task
    """

# message: the message of the cancellation of the gravity task (currently hardcoded to "success")
    message: str = Field(default="")

class CancelDatasetRequest(BaseModel):
    """
     CancelDatasetRequest is the request message for cancelling a dataset build
    """

# dataset_id: the ID of the dataset
    dataset_id: str = Field(default="")

class CancelDatasetResponse(BaseModel):
    """
     CancelDatasetResponse is the response message for cancelling a dataset build
    """

# message: the message of the cancellation of the dataset build (currently hardcoded to "success")
    message: str = Field(default="")

class DatasetBillingCorrectionRequest(BaseModel):
    """
     DatasetBillingCorrectionRequest is the request message for refunding a user
    """

# requested_row_count: number of rows expected by the user
    requested_row_count: int = Field(default=0)
# actual_row_count: number of rows returned by gravity
    actual_row_count: int = Field(default=0)

class DatasetBillingCorrectionResponse(BaseModel):
    """
     DatasetBillingCorrectionResponse is the response message for refunding a user
    """

# refund_amount
    refund_amount: float = Field(default=0.0)



import asyncio
from typing import Dict, List, Union

import grpc

from macrocosmos import __package_name__, __version__
from macrocosmos.generated.gravity.v1 import gravity_p2p, gravity_pb2, gravity_pb2_grpc
from macrocosmos.types import MacrocosmosError
from macrocosmos.resources._client import BaseClient


class AsyncGravity:
    """Asynchronous Gravity resource for the Data Universe (subnet 13) API on Bittensor."""

    def __init__(self, client: BaseClient):
        """
        Initialize the asynchronous Gravity resource.

        Args:
            client: The client to use for the resource.
        """
        self._client = client

    async def GetGravityTasks(
        self,
        gravity_task_id: str = "",
        include_crawlers: bool = False,
    ) -> gravity_pb2.GetGravityTasksResponse:
        """
        List all gravity tasks for a user.

        Args:
            gravity_task_id: The ID of the gravity task (optional, if not provided, all gravity tasks for the user will be returned).
            include_crawlers: Whether to include the crawler states in the response. (default: False)

        Returns:
            A response containing the gravity tasks.
        """
        request = gravity_pb2.GetGravityTasksRequest(
            gravity_task_id=gravity_task_id,
            include_crawlers=include_crawlers,
        )

        return await self._make_request("GetGravityTasks", request)

    async def GetCrawler(
        self,
        crawler_id: str,
    ) -> gravity_pb2.GetCrawlerResponse:
        """
        Get a single crawler by its ID.

        Args:
            crawler_id: The ID of the crawler to get.

        Returns:
            A response containing the crawler details.
        """
        if not crawler_id:
            raise AttributeError("crawler_id is a required parameter")

        request = gravity_pb2.GetCrawlerRequest(crawler_id=crawler_id)

        return await self._make_request("GetCrawler", request)

    async def CreateGravityTask(
        self,
        gravity_tasks: List[Union[gravity_p2p.GravityTask, Dict]] = None,
        name: str = "",
        notification_requests: List[
            Union[gravity_p2p.NotificationRequest, Dict]
        ] = None,
        gravity_task_id: str = "",
    ) -> gravity_pb2.CreateGravityTaskResponse:
        """
        Create a new gravity task.

        Args:
            gravity_tasks: The list of gravity task criteria for the crawlers.
            name: The name of the gravity task (optional).
            notification_requests: The details of the notifications to be sent (optional).
            gravity_task_id: The ID of the gravity task (optional).

        Returns:
            A response containing the ID of the created gravity task.
        """
        proto_gravity_tasks = []
        if gravity_tasks:
            for task in gravity_tasks:
                if isinstance(task, gravity_p2p.GravityTask):
                    proto_gravity_tasks.append(
                        gravity_pb2.GravityTask(**task.model_dump())
                    )
                elif isinstance(task, dict):
                    proto_gravity_tasks.append(gravity_pb2.GravityTask(**task))
                else:
                    raise TypeError(f"Invalid type for gravity task: {type(task)}")
        else:
            raise AttributeError("gravity_tasks is a required parameter")

        proto_notification_requests = []
        if notification_requests:
            for notification in notification_requests:
                if isinstance(notification, gravity_p2p.NotificationRequest):
                    proto_notification_requests.append(
                        gravity_pb2.NotificationRequest(**notification.model_dump())
                    )
                elif isinstance(notification, dict):
                    proto_notification_requests.append(
                        gravity_pb2.NotificationRequest(**notification)
                    )
                else:
                    raise TypeError(
                        f"Invalid type for notification request: {type(notification)}"
                    )

        request = gravity_pb2.CreateGravityTaskRequest(
            gravity_tasks=proto_gravity_tasks,
            name=name,
            notification_requests=proto_notification_requests,
            gravity_task_id=gravity_task_id,
        )

        return await self._make_request("CreateGravityTask", request)

    async def BuildDataset(
        self,
        crawler_id: str,
        max_rows: int,
        notification_requests: List[
            Union[gravity_p2p.NotificationRequest, Dict]
        ] = None,
    ) -> gravity_pb2.BuildDatasetResponse:
        """
        Build a dataset for a single crawler.

        Args:
            crawler_id: The ID of the crawler to build a dataset for.
            max_rows: The maximum number of rows to include in the dataset.
            notification_requests: The details of the notifications to be sent (optional).
        Returns:
            A response containing the dataset that was built.
        """
        if not crawler_id:
            raise AttributeError("crawler_id is a required parameter")

        proto_notification_requests = []
        if notification_requests:
            for notification in notification_requests:
                if isinstance(notification, gravity_p2p.NotificationRequest):
                    proto_notification_requests.append(
                        gravity_pb2.NotificationRequest(**notification.model_dump())
                    )
                elif isinstance(notification, dict):
                    proto_notification_requests.append(
                        gravity_pb2.NotificationRequest(**notification)
                    )
                else:
                    raise TypeError(
                        f"Invalid type for notification request: {type(notification)}"
                    )

        request = gravity_pb2.BuildDatasetRequest(
            crawler_id=crawler_id,
            max_rows=max_rows,
            notification_requests=proto_notification_requests,
        )

        return await self._make_request("BuildDataset", request)

    async def GetDataset(
        self,
        dataset_id: str,
    ) -> gravity_pb2.GetDatasetResponse:
        """
        Get the status of a dataset.

        Args:
            dataset_id: The ID of the dataset to get the status for.

        Returns:
            A response containing the dataset status.
        """
        if not dataset_id:
            raise AttributeError("dataset_id is a required parameter")

        request = gravity_pb2.GetDatasetRequest(
            dataset_id=dataset_id,
        )

        return await self._make_request("GetDataset", request)

    async def CancelGravityTask(
        self,
        gravity_task_id: str,
    ) -> gravity_pb2.CancelGravityTaskResponse:
        """
        Cancel a gravity task.

        Args:
            gravity_task_id: The ID of the gravity task to cancel.

        Returns:
            A response containing the cancellation status.
        """
        if not gravity_task_id:
            raise AttributeError("gravity_task_id is a required parameter")

        request = gravity_pb2.CancelGravityTaskRequest(
            gravity_task_id=gravity_task_id,
        )

        return await self._make_request("CancelGravityTask", request)

    async def CancelDataset(
        self,
        dataset_id: str,
    ) -> gravity_pb2.CancelDatasetResponse:
        """
        Cancel a dataset.

        Args:
            dataset_id: The ID of the dataset to cancel.

        Returns:
            A response containing the cancellation status.
        """
        if not dataset_id:
            raise AttributeError("dataset_id is a required parameter")

        request = gravity_pb2.CancelDatasetRequest(
            dataset_id=dataset_id,
        )

        return await self._make_request("CancelDataset", request)

    async def _make_request(self, method_name, request):
        """
        Make a request to the Gravity service.

        Args:
            method_name: The name of the method to call.
            request: The request message.

        Returns:
            The response from the service.
        """
        metadata = [
            ("x-source", self._client.app_name),
            ("x-client-id", __package_name__),
            ("x-client-version", __version__),
            ("authorization", f"Bearer {self._client.api_key}"),
            (
                "remote-user-email",
                "example@example.com",
            ),  # TODO: Remove this once authentication is fully implemented
        ]

        compression = grpc.Compression.Gzip if self._client.compress else None

        retries = 0
        last_error = None
        while retries <= self._client.max_retries:
            try:
                channel = self._client.get_channel()
                stub = gravity_pb2_grpc.GravityServiceStub(channel)
                method = getattr(stub, method_name)
                response = await method(
                    request,
                    metadata=metadata,
                    timeout=self._client.timeout,
                    compression=compression,
                )
                await channel.close()
                return response
            except grpc.RpcError as e:
                last_error = MacrocosmosError(f"RPC error: {e.code()}: {e.details()}")
                retries += 1
                await channel.close()
            except Exception as e:
                await channel.close()
                raise MacrocosmosError(f"Error calling {method_name}: {e}")

        raise last_error


class SyncGravity:
    """Synchronous Gravity resource for the Data Universe (subnet 13) API on Bittensor."""

    def __init__(self, client: BaseClient):
        """
        Initialize the synchronous Gravity resource.

        Args:
            client: The client to use for the resource.
        """
        self._client = client
        self._async_gravity = AsyncGravity(client)

    def GetGravityTasks(
        self,
        gravity_task_id: str = "",
        include_crawlers: bool = False,
    ) -> gravity_pb2.GetGravityTasksResponse:
        """
        List all gravity tasks for a user synchronously.

        Args:
            gravity_task_id: The ID of the gravity task (optional, if not provided, all gravity tasks for the user will be returned).
            include_crawlers: Whether to include the crawler states in the response.

        Returns:
            A response containing the gravity tasks.
        """
        return asyncio.run(
            self._async_gravity.GetGravityTasks(
                gravity_task_id=gravity_task_id,
                include_crawlers=include_crawlers,
            )
        )

    def CreateGravityTask(
        self,
        gravity_tasks: List[Union[gravity_p2p.GravityTask, Dict]] = None,
        name: str = "",
        notification_requests: List[
            Union[gravity_p2p.NotificationRequest, Dict]
        ] = None,
        gravity_task_id: str = "",
    ) -> gravity_pb2.CreateGravityTaskResponse:
        """
        Create a new gravity task synchronously.

        Args:
            gravity_tasks: The list of gravity task criteria for the crawlers.
            name: The name of the gravity task (optional).
            notification_requests: The details of the notifications to be sent (optional).
            gravity_task_id: The ID of the gravity task (optional).

        Returns:
            A response containing the ID of the created gravity task.
        """
        return asyncio.run(
            self._async_gravity.CreateGravityTask(
                gravity_tasks=gravity_tasks,
                name=name,
                notification_requests=notification_requests,
                gravity_task_id=gravity_task_id,
            )
        )

    def BuildDataset(
        self,
        crawler_id: str,
        max_rows: int,
        notification_requests: List[
            Union[gravity_p2p.NotificationRequest, Dict]
        ] = None,
    ) -> gravity_pb2.BuildDatasetResponse:
        """
        Build a dataset for a single crawler synchronously.

        Args:
            crawler_id: The ID of the crawler to build a dataset for.
            max_rows: The maximum number of rows to include in the dataset.
            notification_requests: The details of the notifications to be sent (optional).

        Returns:
            A response containing the dataset that was built.
        """
        return asyncio.run(
            self._async_gravity.BuildDataset(
                crawler_id=crawler_id,
                max_rows=max_rows,
                notification_requests=notification_requests,
            )
        )

    def GetDataset(
        self,
        dataset_id: str,
    ) -> gravity_pb2.GetDatasetResponse:
        """
        Get the status of a dataset build synchronously.

        Args:
            dataset_id: The ID of the dataset to get the status for.

        Returns:
            A response containing the dataset status.
        """
        return asyncio.run(
            self._async_gravity.GetDataset(
                dataset_id=dataset_id,
            )
        )

    def CancelGravityTask(
        self,
        gravity_task_id: str,
    ) -> gravity_pb2.CancelGravityTaskResponse:
        """
        Cancel a gravity task synchronously.

        Args:
            gravity_task_id: The ID of the gravity task to cancel.

        Returns:
            A response containing the cancellation status.
        """
        return asyncio.run(
            self._async_gravity.CancelGravityTask(
                gravity_task_id=gravity_task_id,
            )
        )

    def CancelDataset(
        self,
        dataset_id: str,
    ) -> gravity_pb2.CancelDatasetResponse:
        """
        Cancel a dataset build synchronously.
        """
        return asyncio.run(
            self._async_gravity.CancelDataset(
                dataset_id=dataset_id,
            )
        )



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings
from macrocosmos.generated.gravity.v1 import gravity_pb2 as gravity_dot_v1_dot_gravity__pb2

GRPC_GENERATED_VERSION = '1.71.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in gravity/v1/gravity_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class GravityServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetGravityTasks = channel.unary_unary(
                '/gravity.v1.GravityService/GetGravityTasks',
                request_serializer=gravity_dot_v1_dot_gravity__pb2.GetGravityTasksRequest.SerializeToString,
                response_deserializer=gravity_dot_v1_dot_gravity__pb2.GetGravityTasksResponse.FromString,
                _registered_method=True)
        self.GetCrawler = channel.unary_unary(
                '/gravity.v1.GravityService/GetCrawler',
                request_serializer=gravity_dot_v1_dot_gravity__pb2.GetCrawlerRequest.SerializeToString,
                response_deserializer=gravity_dot_v1_dot_gravity__pb2.GetCrawlerResponse.FromString,
                _registered_method=True)
        self.CreateGravityTask = channel.unary_unary(
                '/gravity.v1.GravityService/CreateGravityTask',
                request_serializer=gravity_dot_v1_dot_gravity__pb2.CreateGravityTaskRequest.SerializeToString,
                response_deserializer=gravity_dot_v1_dot_gravity__pb2.CreateGravityTaskResponse.FromString,
                _registered_method=True)
        self.BuildDataset = channel.unary_unary(
                '/gravity.v1.GravityService/BuildDataset',
                request_serializer=gravity_dot_v1_dot_gravity__pb2.BuildDatasetRequest.SerializeToString,
                response_deserializer=gravity_dot_v1_dot_gravity__pb2.BuildDatasetResponse.FromString,
                _registered_method=True)
        self.GetDataset = channel.unary_unary(
                '/gravity.v1.GravityService/GetDataset',
                request_serializer=gravity_dot_v1_dot_gravity__pb2.GetDatasetRequest.SerializeToString,
                response_deserializer=gravity_dot_v1_dot_gravity__pb2.GetDatasetResponse.FromString,
                _registered_method=True)
        self.CancelGravityTask = channel.unary_unary(
                '/gravity.v1.GravityService/CancelGravityTask',
                request_serializer=gravity_dot_v1_dot_gravity__pb2.CancelGravityTaskRequest.SerializeToString,
                response_deserializer=gravity_dot_v1_dot_gravity__pb2.CancelGravityTaskResponse.FromString,
                _registered_method=True)
        self.CancelDataset = channel.unary_unary(
                '/gravity.v1.GravityService/CancelDataset',
                request_serializer=gravity_dot_v1_dot_gravity__pb2.CancelDatasetRequest.SerializeToString,
                response_deserializer=gravity_dot_v1_dot_gravity__pb2.CancelDatasetResponse.FromString,
                _registered_method=True)
        self.DatasetBillingCorrection = channel.unary_unary(
                '/gravity.v1.GravityService/DatasetBillingCorrection',
                request_serializer=gravity_dot_v1_dot_gravity__pb2.DatasetBillingCorrectionRequest.SerializeToString,
                response_deserializer=gravity_dot_v1_dot_gravity__pb2.DatasetBillingCorrectionResponse.FromString,
                _registered_method=True)


class GravityServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetGravityTasks(self, request, context):
        """Lists all data collection tasks for a user
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetCrawler(self, request, context):
        """Get a single crawler by its ID
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateGravityTask(self, request, context):
        """Create a new gravity task
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def BuildDataset(self, request, context):
        """Build a dataset for a single crawler
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetDataset(self, request, context):
        """Get the dataset build status and results
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CancelGravityTask(self, request, context):
        """Cancel a gravity task and any crawlers associated with it
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CancelDataset(self, request, context):
        """Cancel dataset build if it is in progress and purges the dataset
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DatasetBillingCorrection(self, request, context):
        """Refund user if fewer rows are returned
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GravityServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetGravityTasks': grpc.unary_unary_rpc_method_handler(
                    servicer.GetGravityTasks,
                    request_deserializer=gravity_dot_v1_dot_gravity__pb2.GetGravityTasksRequest.FromString,
                    response_serializer=gravity_dot_v1_dot_gravity__pb2.GetGravityTasksResponse.SerializeToString,
            ),
            'GetCrawler': grpc.unary_unary_rpc_method_handler(
                    servicer.GetCrawler,
                    request_deserializer=gravity_dot_v1_dot_gravity__pb2.GetCrawlerRequest.FromString,
                    response_serializer=gravity_dot_v1_dot_gravity__pb2.GetCrawlerResponse.SerializeToString,
            ),
            'CreateGravityTask': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateGravityTask,
                    request_deserializer=gravity_dot_v1_dot_gravity__pb2.CreateGravityTaskRequest.FromString,
                    response_serializer=gravity_dot_v1_dot_gravity__pb2.CreateGravityTaskResponse.SerializeToString,
            ),
            'BuildDataset': grpc.unary_unary_rpc_method_handler(
                    servicer.BuildDataset,
                    request_deserializer=gravity_dot_v1_dot_gravity__pb2.BuildDatasetRequest.FromString,
                    response_serializer=gravity_dot_v1_dot_gravity__pb2.BuildDatasetResponse.SerializeToString,
            ),
            'GetDataset': grpc.unary_unary_rpc_method_handler(
                    servicer.GetDataset,
                    request_deserializer=gravity_dot_v1_dot_gravity__pb2.GetDatasetRequest.FromString,
                    response_serializer=gravity_dot_v1_dot_gravity__pb2.GetDatasetResponse.SerializeToString,
            ),
            'CancelGravityTask': grpc.unary_unary_rpc_method_handler(
                    servicer.CancelGravityTask,
                    request_deserializer=gravity_dot_v1_dot_gravity__pb2.CancelGravityTaskRequest.FromString,
                    response_serializer=gravity_dot_v1_dot_gravity__pb2.CancelGravityTaskResponse.SerializeToString,
            ),
            'CancelDataset': grpc.unary_unary_rpc_method_handler(
                    servicer.CancelDataset,
                    request_deserializer=gravity_dot_v1_dot_gravity__pb2.CancelDatasetRequest.FromString,
                    response_serializer=gravity_dot_v1_dot_gravity__pb2.CancelDatasetResponse.SerializeToString,
            ),
            'DatasetBillingCorrection': grpc.unary_unary_rpc_method_handler(
                    servicer.DatasetBillingCorrection,
                    request_deserializer=gravity_dot_v1_dot_gravity__pb2.DatasetBillingCorrectionRequest.FromString,
                    response_serializer=gravity_dot_v1_dot_gravity__pb2.DatasetBillingCorrectionResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'gravity.v1.GravityService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('gravity.v1.GravityService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class GravityService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetGravityTasks(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gravity.v1.GravityService/GetGravityTasks',
            gravity_dot_v1_dot_gravity__pb2.GetGravityTasksRequest.SerializeToString,
            gravity_dot_v1_dot_gravity__pb2.GetGravityTasksResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetCrawler(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gravity.v1.GravityService/GetCrawler',
            gravity_dot_v1_dot_gravity__pb2.GetCrawlerRequest.SerializeToString,
            gravity_dot_v1_dot_gravity__pb2.GetCrawlerResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CreateGravityTask(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gravity.v1.GravityService/CreateGravityTask',
            gravity_dot_v1_dot_gravity__pb2.CreateGravityTaskRequest.SerializeToString,
            gravity_dot_v1_dot_gravity__pb2.CreateGravityTaskResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def BuildDataset(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gravity.v1.GravityService/BuildDataset',
            gravity_dot_v1_dot_gravity__pb2.BuildDatasetRequest.SerializeToString,
            gravity_dot_v1_dot_gravity__pb2.BuildDatasetResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetDataset(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gravity.v1.GravityService/GetDataset',
            gravity_dot_v1_dot_gravity__pb2.GetDatasetRequest.SerializeToString,
            gravity_dot_v1_dot_gravity__pb2.GetDatasetResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CancelGravityTask(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gravity.v1.GravityService/CancelGravityTask',
            gravity_dot_v1_dot_gravity__pb2.CancelGravityTaskRequest.SerializeToString,
            gravity_dot_v1_dot_gravity__pb2.CancelGravityTaskResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CancelDataset(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gravity.v1.GravityService/CancelDataset',
            gravity_dot_v1_dot_gravity__pb2.CancelDatasetRequest.SerializeToString,
            gravity_dot_v1_dot_gravity__pb2.CancelDatasetResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DatasetBillingCorrection(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gravity.v1.GravityService/DatasetBillingCorrection',
            gravity_dot_v1_dot_gravity__pb2.DatasetBillingCorrectionRequest.SerializeToString,
            gravity_dot_v1_dot_gravity__pb2.DatasetBillingCorrectionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)
